@@ -315,14 +315,20 @@ let invariant env (x, a) =
   let env = { env with variables = Env.add env.variables x () } in
   x, assertion env a, []
 
+let own_region_def env (id, ce) =
+  let id = Ident.create id in
+  let ce = Opt.map (class_expr env) ce in
+  Opt.iter (Regiontypes.add id) ce;
+  id, ce
+
 let class_def env x =
   let rp = List.map Ident.create x.P.c_region_params in
-  let ro = List.map Ident.create x.P.c_owned_regions in
+  let ro = List.map (own_region_def env) x.P.c_owned_regions in
   let tp = List.map Ident.create x.P.c_type_params in
   let name = Ident.create x.P.c_name in
   let env = { env with classes = Env.add env.classes name () } in
   let iregions = add_list env.regions rp in
-  let iregions = add_list iregions ro in
+  let iregions = add_list iregions (List.map fst ro) in
   let ienv =
     { env with
         regions = iregions;
@@ -336,21 +342,26 @@ let class_def env x =
     B.c_type = typ ienv x.P.c_type;
     B.c_invariant = invariant ienv x.P.c_invariant }
 
-let rec add_rvar acc r =
+let rec add_rvar env acc r =
   match r.node with
-    | P.RVar s -> StringSet.add s acc
-    | P.RSub (r, _) -> add_rvar acc r
+    | P.RVar s ->
+        if Env.mem_string env.regions s then
+          acc
+        else
+          StringSet.add s acc
+    | P.RSub (r, _) ->
+        add_rvar env acc r
 
-let rec free_regions_of_type_expr acc x =
+let rec free_regions_of_type_expr env acc x =
   match x.node with
     | P.TEPointer ((rl, tl, _), r) ->
-        let acc = Opt.fold add_rvar acc r in
-        let acc = List.fold_left add_rvar acc rl in
-        List.fold_left free_regions_of_type_expr acc tl
+        let acc = Opt.fold (add_rvar env) acc r in
+        let acc = List.fold_left (add_rvar env) acc rl in
+        List.fold_left (free_regions_of_type_expr env) acc tl
     | P.TETuple tl ->
-        List.fold_left free_regions_of_type_expr acc tl
+        List.fold_left (free_regions_of_type_expr env) acc tl
     | P.TESum (a, b) ->
-        free_regions_of_type_expr (free_regions_of_type_expr acc a) b
+        free_regions_of_type_expr env (free_regions_of_type_expr env acc a) b
 
 let rec free_variables_of_type_expr env acc x =
   match x.node with
@@ -394,7 +405,7 @@ let rec free_type_variables_of_assertion env acc x =
         free_type_variables_of_assertion env
           (free_variables_of_type_expr env acc t) a
 
-let rec free_region_variables_of_assertion acc x =
+let rec free_region_variables_of_assertion env acc x =
   match x.node with
     | P.PTrue
     | P.PFalse ->
@@ -403,11 +414,11 @@ let rec free_region_variables_of_assertion acc x =
     | P.PImplies (a, b)
     | P.PAnd (a, b)
     | P.POr (a, b) ->
-        free_region_variables_of_assertion
-          (free_region_variables_of_assertion acc a) b
+        free_region_variables_of_assertion env
+          (free_region_variables_of_assertion env acc a) b
     | P.POld a
     | P.PNot a ->
-        free_region_variables_of_assertion acc a
+        free_region_variables_of_assertion env acc a
     | P.PEqual _
     | P.PDiff _
     | P.PLt _
@@ -419,8 +430,8 @@ let rec free_region_variables_of_assertion acc x =
         acc
     | P.PForall (_, t, a)
     | P.PExists (_, t, a) ->
-        free_region_variables_of_assertion
-          (free_regions_of_type_expr acc t) a
+        free_region_variables_of_assertion env
+          (free_regions_of_type_expr env acc t) a
 
 let type_and_region_variables fr ft l =
   let regions =
@@ -431,6 +442,8 @@ let type_and_region_variables fr ft l =
     |> StringSet.elements
     |> List.map Ident.create
   in
+  log "Created identifiers %a@." (Pp.list ~sep: Pp.comma_space Pp.identifier)
+    regions;
   let type_vars =
     List.fold_left
       ft
@@ -451,7 +464,7 @@ let enter_type_and_region (t, r) env =
 let type_and_region_variables_of_types env types =
   enter_type_and_region
     (type_and_region_variables
-       free_regions_of_type_expr
+       (free_regions_of_type_expr env)
        (free_variables_of_type_expr env)
        types)
     env
@@ -459,7 +472,7 @@ let type_and_region_variables_of_types env types =
 let type_and_region_variables_of_assertion env body =
   enter_type_and_region
     (type_and_region_variables
-       free_region_variables_of_assertion
+       (free_region_variables_of_assertion env)
        (free_type_variables_of_assertion env)
        [ body ])
     env
@@ -523,6 +536,7 @@ let axiom_def env x =
     type_and_region_variables_of_assertion env x.P.ax_assertion in
   env,
   { B.ax_name = Ident.create x.P.ax_name;
+    B.ax_is_lemma = x.P.ax_is_lemma;
     B.ax_assertion = assertion ienv x.P.ax_assertion;
     B.ax_type_variables = type_vars;
     B.ax_region_variables = regions }
@@ -545,6 +559,9 @@ let predicate_def env x =
     List.map (fun (s, t) -> Ident.create s, typ ienv t) x.P.p_params in
   let ienv =
     { ienv with variables = add_list env.variables (List.map fst params) } in
+  log "Region variables of predicate %a are %a@."
+    Pp.identifier name
+    (Pp.list ~sep: Pp.comma_space Pp.identifier) regions;
   { env with predicates = Env.add env.predicates name () },
   { B.p_name = name;
     B.p_params = params;
