@@ -30,6 +30,48 @@
 
 (** Why interpretation of Capucine programs. *)
 
+(*
+How regions are translated:
+  There are 3 references (R when group, V for value and P for pointer when
+  singleton) which may actually sometimes not be references. Moreover,
+  the fact that there is a region tree and region substitutions when
+  calling complicates stuff a bit.
+
+It depends on how the region is bound.
+  1) The region is bound using the region ... in operation. The three references
+     are created.
+  2) The region is bound at the level of the function.
+    a) No parent* of the region has a positive permission
+       (read-only: only a group version of the region, with no reference).
+    b) A parent* has a positive permission.
+       (read-write: all three references are created)
+  3) In logic predicates (predicate definitions, axioms and invariants),
+     all regions are considered group read-only regions.
+
+So, when calling a predicate, one must find a group region (not a reference).
+Depending on how the region is bound you may either already have it,
+already have it but under a reference, or you have to build it from the
+singleton references, or if the region is being focused you have to
+use an if-then-else.
+
+Finally, when calling a function, one first have to find out whether the
+region is read-only or not in the called function. If read-only, then its
+as if calling a predicate, and if not, then the three references are
+available anyway in the caller.
+
+Finding out whether a region is read-only has a trap because of the region
+tree: if A is read-only, A.B may not be. To find out whether R.r is read-only,
+first look for permissions on R.r, then for permissions on R, and so on.
+When calling a function, these have to be looked both in the callee and
+in the caller.
+
+Another trap when checking whether a region is read-only appears with
+function calls where some regions are unified. If a non-read-only region A
+and a read-only region B are unified, then something bad happens. This
+actually may happen in real programs if B is later unified with A.B or
+something (see an older version of subobs).
+*)
+
 open Typing_common
 open Ast
 open Bast
@@ -94,23 +136,37 @@ let rec set_of_region_tree ?(acc = RegionSet.empty) r =
   List.fold_left
     (fun acc id -> set_of_region_tree ~acc (RSub (r, Ident.name id)))
     (RegionSet.add r acc)
-    (class_of_region r).c_owned_regions
+    (List.map fst (class_of_region r).c_owned_regions)
 
 let list_of_region_tree r =
   let set = set_of_region_tree r in
   let l = RegionSet.elements set in
   List.sort Region.compare l
 
+(* This function is used to determine whether [region] is read-only according
+   to permissions [perms]. A region is not read-only if it or one of its
+   transitive owner has a permission on it. *)
 let is_region_root_positive_in perms region =
+  let rec an_owner_is r o =
+    Infer.eq_region r o ||
+      match o with
+        | RVar _ -> false
+        | RSub (o', _) -> an_owner_is r o'
+  in
   List.exists
     (function
        | PSub _ ->
            false
        | PEmpty r | PArrow (_, r) | PGroup r | POpen r | PClosed r ->
-           root_of_region (Unify.expand_region r)
-           = root_of_region (Unify.expand_region region))
+           let r = Unify.expand_region r in
+           let region = Unify.expand_region region in
+           (* return whether an owner of region is r *)
+           an_owner_is r region)
+(*           root_of_region (Unify.expand_region r)
+           = root_of_region (Unify.expand_region region))*)
     perms
 
+(* for debugging only *)
 let is_region_root_positive_in p r =
   let res = is_region_root_positive_in p r in
   log "is_region_root_positive_in %a %a = %b@."
@@ -130,6 +186,8 @@ let is_singleton_in perms region =
 
 type region_variable = G | P | V
 
+(* (deref_of_perms perms) computes whether a region should be dereferenced
+   to read into it. It is used in Deref expressions and terms. *)
 let deref_of_perms = is_region_root_positive_in
 let always_deref _ = true
 let never_deref _ = false
@@ -238,16 +296,39 @@ let region_arguments_of_region perms r =
 
 (* when calling values (must be compatible with the above function) *)
 let region_arguments_of_region_app ~perms ~consumes r =
+  log "@[<hv 2>REGION_ARGUMENTS_OF_REGION_APP@ @[<hv 2>perms =@ %a@]@ @[<hv 2>consumes =@ %a@]@ @[r = %a@]@]@."
+    (Pp.list ~sep: Pp.comma_space Pp.permission) perms
+    (Pp.list ~sep: Pp.comma_space Pp.permission) consumes
+    Pp.region r;
   if is_region_root_positive_in consumes r then
-    List.map var (List.map fst (region_references_of_region r))
-  else
-    if is_singleton_in perms r then
-      [ expr_set
-          expr_empty
-          (read_region in_expr always_deref r P)
-          (read_region in_expr always_deref r V) ]
-    else
-      [ read_region in_expr always_deref r G ]
+    (log "==> I'll use the three references because the callee needs so.@.";
+     List.map var (List.map fst (region_references_of_region r)))
+  else begin
+    match Infer.find_deref_kind r perms with
+      | DKSingleton ->
+          log "==> I'll construct a non-ref map from a singleton.@.";
+          [ expr_set
+              expr_empty
+              (read_region in_expr always_deref r P)
+              (read_region in_expr always_deref r V) ]
+      | DKUnavailable ->
+          log "==> No permission is available, but maybe an owner is?@.";
+          if is_region_root_positive_in perms r then begin
+            log "==> Found an owner: region is not read-only yet, \
+I'll deref the map.@.";
+            [ read_region in_expr always_deref r G ]
+          end else begin
+            log "==> No owner found: region is already read-only, \
+I'll use the non-ref map.@.";
+            [ read_region in_expr never_deref r G ]
+          end
+      | DKGroup ->
+          log "==> Region is group, I'll convert the ref into a read-only.@.";
+          [ read_region in_expr always_deref r G ]
+      | DKFocus s ->
+          (* TODO *)
+          assert false
+  end
 
 let region_arguments_of_value x =
   call_regions x
@@ -471,7 +552,8 @@ let rec expr x =
     | If (a, b, c) ->
         expr_if (expr a) (expr b) (expr c)
     | While (a, b, i, perms) ->
-        loop ~invariant: (assertion perms false i) (expr a) (expr b)
+        loop ~invariant: 
+          (assertion perms (Some at_init) None i) (expr a) (expr b)
     | Assign (a, b, r) ->
         expr_ignore
           (expr a)
@@ -574,13 +656,49 @@ let rec expr x =
           copy_region_trees s r;
         ]
     | AdoptRegion (s, r, AKSingletonToGroup) ->
+        let lbl = Ident.create "before_adopt" in
+        let new_g_value =
+          a_eq
+            (read_region in_term always_deref r G)
+            (set
+               (read_region ~at: lbl in_term always_deref r G)
+               (read_region in_term always_deref s P)
+               (read_region in_term always_deref s V))
+        in
+        let new_g_extends =
+          (* La version commentée semble aider les prouveurs, mais la version
+             utilisée donne plus d'information (on pourrait mettre les deux
+             mais ça ferait beaucoup d'hypothèses) *)
+(*          extends
+            (read_region in_term always_deref r G)
+            (read_region ~at: lbl in_term always_deref r G)*)
+          a_not
+            (in_region
+               (read_region in_term always_deref s P)
+               (read_region ~at: lbl in_term always_deref r G))
+        in
+        let p = Ident.create "p" in
+        let pointer_is_different =
+          a_forall p pointer
+            (a_implies
+               (in_region (term_var p) (read_region ~at: lbl in_term always_deref r G))
+               (a_neq (term_var p) (read_region in_term always_deref s P)))
+        in
         seq [
-          assign
+(*          assign
             (region_var r G)
             (expr_set
                (read_region in_expr always_deref r G)
                (read_region in_expr always_deref s P)
-               (read_region in_expr always_deref s V));
+               (read_region in_expr always_deref s V)); *)
+          label lbl
+            (black_box
+               (type_annot
+                  ~writes: [ region_var r G ]
+                  ~post: (a_and (a_and new_g_value new_g_extends) pointer_is_different)
+                  (type_base unit)));
+          (* TODO: faire les extends aussi pour les sous-regions et dans les
+             autres cas d'adoption (mais pas tous !) *)
           append_region_trees s r;
         ]
     | AdoptRegion (s, r, AKGroupToGroup) ->
@@ -617,6 +735,11 @@ let rec expr x =
            into t *)
         (* TODO: cas d'un unfocus de groupe ? *)
         seq [
+          (* TODO: commenter ce assume dans ma thèse *)
+          assume
+            (in_region
+               (read_region in_term always_deref s P)
+               (read_region in_term always_deref r G));
           assign
             (region_var r G)
             (expr_set
@@ -658,20 +781,19 @@ let rec expr x =
     | Typed (e, _) ->
         expr e
     | Assert (a, perms) ->
-        expr_assert (assertion perms false a)
+        expr_assert (assertion perms (Some at_init) None a)
     | Old _ ->
         Loc.locate_error x.loc
           "``old'' makes no sense in expressions, it can only be used in terms"
 
-and term perms old t =
+and term perms old current t =
 (*
   let is_singleton = is_singleton_in perms in
 *)
-  let old_term t = term perms true t in
-  let term t = term perms old t in
-  let at = if old then Some at_old else None in
-  let logic_app_region_arguments = logic_app_region_arguments ?at in
-  let read_region = read_region ?at in
+  let old_term t = term perms old old t in
+  let term t = term perms old current t in
+  let logic_app_region_arguments = logic_app_region_arguments ?at:current in
+  let read_region = read_region ?at: current in
   match t.node with
     | Const CUnit ->
         term_void
@@ -682,7 +804,8 @@ and term perms old t =
     | Tuple l ->
         term_tuple (List.map term l)
     | LogicApp (x, l, rs, _) ->
-        log "LogicApp(%a): old = %b@." Pp.identifier x old;
+        log "LogicApp(%a): old = %a@." Pp.identifier x 
+          (Pp.option Pp.identifier) old;
         let regions =
           logic_app_region_arguments in_term perms rs in
         term_app x (regions @ List.map term l)
@@ -768,12 +891,11 @@ and term perms old t =
         log "term = %a@." Pp.expr t;
         assert false (* impossible *)
 
-and assertion perms old a =
-  let old_assertion a = assertion perms true a in
-  let assertion a = assertion perms old a in
-  let term t = term perms false t in
-  let at = if old then Some at_old else None in
-  let logic_app_region_arguments = logic_app_region_arguments ?at in
+and assertion perms old current a =
+  let old_assertion a = assertion perms old old a in
+  let assertion a = assertion perms old current a in
+  let term t = term perms old current t in
+  let logic_app_region_arguments = logic_app_region_arguments ?at:current in
   match a.node with
     | PTrue ->
         a_true
@@ -868,11 +990,11 @@ let let_of_value pre post args x body =
        ~pre
        ~post
        ~args
-       (seq [
+       (label at_init (seq [
           assume_singletons x.v_params x.v_consumes;
           assume_permissions x.v_consumes;
           expr body;
-        ]))
+        ])))
 
 (* if the return type is a pointer type of a singleton region, we add the
    following precondition: result = R_p *)
@@ -890,8 +1012,8 @@ let add_pointer_equality x post =
 
 let value_def acc x =
   Effects.fixpoint [x];
-  let pre = assertion x.v_consumes false x.v_pre in
-  let post = assertion x.v_produces false x.v_post in
+  let pre = assertion x.v_consumes (Some at_old) None x.v_pre in
+  let post = assertion x.v_produces (Some at_old) None x.v_post in
   let post = add_pointer_equality x post in
   let args = arguments_of_value x in
   let acc = parameter_of_value pre post args x :: acc in
@@ -913,7 +1035,7 @@ let invariant_predicate_declaration_of c =
   decl_predicate
     (invariant_id c)
     (regs @ [ x, translate_type c.c_type ])
-    (Some (assertion [] false a))
+    (Some (assertion [] None None a))
     (* no permission for invariants, as permissions are used to test whether
        a region is singleton, but the only regions which can be read in
        invariants are the owned regions, which are group *)
@@ -925,26 +1047,77 @@ let logic_type_def acc x =
   let lt = decl_logic_type ~args: x.lt_type_params x.lt_name in
   lt :: acc
 
+(* name: name of the logic function or predicate (as an Ident.t)
+   eq: equality (for logic functions) or equivalence (for predicates)
+   app: application (a_app for predicates, term_app for logic functions)
+   regions: names of the region parameters (as Ident.t), and their types
+     (already translated)
+   param_types: types of the normal parameters (already translated) *)
+let auto_footprint_axiom eq app name regions params =
+  let axiom_id = Ident.create (Ident.name name ^ "_footprint") in
+  let param_names = List.map fst params in
+  let region_names = List.map fst regions in
+  let extended_regions =
+    List.map
+      (fun (id, ty) -> Ident.create (Ident.name id ^ "_x"), ty)
+      regions
+  in
+  let extended_region_names = List.map fst extended_regions in
+  let body =
+    eq
+      (app name (List.map term_var (extended_region_names @ param_names)))
+      (app name (List.map term_var (region_names @ param_names)))
+  in
+  let body =
+    List.fold_left2
+      (fun acc r' r -> a_implies (extends (term_var r') (term_var r)) acc)
+      body
+      extended_region_names
+      region_names
+  in
+  let body =
+    List.fold_left
+      (fun acc (id, ty) -> a_forall id ty acc)
+      body
+      (extended_regions @ regions @ params)
+  in
+  decl_axiom
+    ~is_lemma: false
+    axiom_id
+    body
+
 let logic_function_def acc x =
-  let regions = region_arguments_of_logic x |> List.map snd in
-  let args = regions @ List.map translate_type x.lf_params in
+  let regions = region_arguments_of_logic x in
+  let region_types = List.map snd regions in
+  let param_types = List.map translate_type x.lf_params in
+  let args = region_types @ param_types in
+  let params =
+    List.map
+      (fun ty -> Ident.create "x", ty)
+      param_types
+  in
   let lf =
     decl_logic
       x.lf_name
       ~args
       (translate_type x.lf_return_type)
   in
-  lf :: acc
+  let acc = lf :: acc in
+  if regions <> [] then
+    auto_footprint_axiom a_eq term_app x.lf_name regions params :: acc
+  else
+    acc
 
 let axiom_def acc x =
   let body =
     List.fold_left
       (fun acc (x, t) -> a_forall x t acc)
-      (assertion [] false x.ax_assertion)
+      (assertion [] None None x.ax_assertion)
       (* no permission, as all regions are assumed group *)
       (region_arguments_of_axiom x)
   in
-  let ax = decl_axiom x.ax_name body in
+  let ax = decl_axiom ~is_lemma:x.ax_is_lemma x.ax_name body 
+  in
   ax :: acc
 
 let predicate_def acc x =
@@ -954,10 +1127,14 @@ let predicate_def acc x =
     decl_predicate
       x.p_name
       (regions @ params)
-      (Opt.map (assertion [] false) x.p_body)
+      (Opt.map (assertion [] None None) x.p_body)
   in
   (* no permission, as all regions are assumed group *)
-  p :: acc
+  let acc = p :: acc in
+  if regions <> [] then
+    auto_footprint_axiom a_iff a_app x.p_name regions params :: acc
+  else
+    acc
 
 let def acc = function
   | Class x -> class_def acc x
