@@ -77,6 +77,9 @@ but is here applied to %d" Pp.identifier lt.lt_name elen len;
 
 (** Insert a black box at the end of the expression, requiring that
     region [s] is a subregion of [r]. *)
+(* One could think that the adoption should be done before. But that is not
+   the case. We type an expression, it is well-typed, and now we want to
+   use the expression with another type, so we need to do some adoptions. *)
 let apply_constraint x (s, r) =
   let p = [ PSub (s, r) ] in
   let bb = { x with node = BlackBox (p, p) } in
@@ -99,19 +102,19 @@ let op = function
   | #int_op -> tbase TInt
   | #bool_op -> tbase TBool
 
-let check_pointer t =
+let check_pointer loc t =
   match t.node with
     | TEPointer _ ->
         ()
     | _ ->
-        Loc.locate_error t.loc "this should be a pointer"
+        Loc.locate_error loc "this should be a pointer"
 
-let class_expr t =
+let class_expr loc t =
   match t.node with
     | TEPointer (ce, _) ->
         ce
     | _ ->
-        Loc.locate_error t.loc "this should be a pointer"
+        Loc.locate_error loc "this should be a pointer"
 
 let rec instanciate_region rs = function
   | RVar id -> List.assoc id rs
@@ -182,18 +185,20 @@ let instanciate_predicate p =
   List.map instanciate_type (List.map snd p.p_params),
   List.map (fun x -> List.assoc x rs) p.p_region_variables
 
-let region_of t =
+let region_of loc t =
   match t.node with
     | TEPointer (_, r) ->
         r
     | _ ->
-        Loc.locate_error t.loc "this should be a pointer"
+        Loc.locate_error loc "this should be a pointer"
 
 let add_list = Bind.add_list
 let add_list2 env l = List.fold_left (fun a (b, c) -> Env.add a b c) env l
 
 (** Return [(t, e)] where [t] is the type and [e] the modified expression. *)
-let rec expr env x =
+let rec expr actually_a_term env x =
+  let expr x = expr actually_a_term x in
+  let expr_as x = expr_as actually_a_term x in
   let return typ node = typ, { node = node; loc = x.loc } in
   match x.node with
     | Const c ->
@@ -276,38 +281,38 @@ but is here applied to %d" Pp.identifier lf.lf_name elen len;
         return (tbase TUnit) (While (a, b, i, []))
     | Assign (a, b, _) ->
         let ta, a = expr env a in
-        let b = expr_as env b (pointed_type env ta) in
-        return (tbase TUnit) (Assign (a, b, region_of ta))
+        let b = expr_as env b (pointed_type x.loc env ta) in
+        return (tbase TUnit) (Assign (a, b, region_of a.loc ta))
     | Deref (a, _, _, _) ->
         let ta, a = expr env a in
         return
-          (pointed_type env ta)
-          (Deref (a, region_of ta, DKUnavailable, []))
+          (pointed_type x.loc env ta)
+          (Deref (a, region_of a.loc ta, DKUnavailable, []))
     | New ((rl, tl, v), r) ->
         let t = tpointer rl tl v r in
         check_type env t;
         return t (New ((rl, tl, v), r))
     | Pack (a, _) ->
         let ta, a = expr env a in
-        check_pointer ta;
-        return (tbase TUnit) (Pack (a, region_of ta))
+        check_pointer a.loc ta;
+        return (tbase TUnit) (Pack (a, region_of a.loc ta))
     | Unpack (a, _) ->
         let ta, a = expr env a in
-        check_pointer ta;
-        return (tbase TUnit) (Unpack (a, region_of ta))
+        check_pointer a.loc ta;
+        return (tbase TUnit) (Unpack (a, region_of a.loc ta))
     | Adopt (a, r, _, ak) ->
         let ta, a = expr env a in
-        let rl, tl, c = class_expr ta in
-        return (tpointer rl tl c r) (Adopt (a, r, region_of ta, ak))
+        let rl, tl, c = class_expr a.loc ta in
+        return (tpointer rl tl c r) (Adopt (a, r, region_of a.loc ta, ak))
     | Focus (a, r, _) ->
         let ta, a = expr env a in
-        let rl, tl, c = class_expr ta in
+        let rl, tl, c = class_expr a.loc ta in
         Regiontypes.add_if_var r (rl, tl, c);
-        return (tpointer rl tl c r) (Focus (a, r, region_of ta))
+        return (tpointer rl tl c r) (Focus (a, r, region_of a.loc ta))
     | Unfocus (a, r, _) ->
         let ta, a = expr env a in
-        check_pointer ta;
-        return (tbase TUnit) (Unfocus (a, r, region_of ta))
+        check_pointer a.loc ta;
+        return (tbase TUnit) (Unfocus (a, r, region_of a.loc ta))
     | Region (v, a, ceo) ->
         let env = { env with regions = Env.add env.regions v () } in
         let ta, a = expr env a in
@@ -336,17 +341,20 @@ but is here applied to %d" Pp.identifier lf.lf_name elen len;
         let ta, a = expr env a in
         return ta (Old a)        
 
-and expr_as env x et =
+and expr_as actually_a_term env x et =
   match x.node with
     | Region (r, a, t) ->
-        { x with node = Region (r, expr_as env a et, t) }
+        { x with node = Region (r, expr_as actually_a_term env a et, t) }
     | _ ->
-        let t, x = expr env x in
+        let t, x = expr actually_a_term env x in
         let constraints = Unify.unify_types x.loc true t et in
-        List.fold_left apply_constraint x constraints
+        if actually_a_term then
+          x
+        else
+          List.fold_left apply_constraint x constraints
 
 and term_as env t ty =
-  expr_as env t ty
+  expr_as true env t ty
 
 and term env t =
   let ty = fresh_utv "term_type" in
@@ -410,6 +418,9 @@ but is here applied to %d" Pp.identifier p.p_name elen len;
   in
   { a with node = node }
 
+let expr_as = expr_as false
+let expr = expr false
+
 let invariant env (x, a, _) t =
   let env = { env with variables = Env.add env.variables x t } in
   let a = assertion env a in
@@ -473,6 +484,7 @@ let axiom_def env x =
   { x with ax_assertion = assertion ienv x.ax_assertion }
 
 let predicate_def env x =
+  log "Typing predicate %a@." Pp.identifier x.p_name;
   let ienv =
     { env with
         type_variables = add_list env.type_variables x.p_type_variables;
