@@ -42,6 +42,7 @@ type logic_type_normal = {
 
 and logic_type =
   | LTNormal of logic_type_normal
+  | LTTuple of logic_type list
   | LTVar of Ident.t
 
 let logic_type ?(args = []) name =
@@ -55,43 +56,15 @@ let int = logic_type (Ident.create "int")
 let bool = logic_type (Ident.create "bool")
 let string = logic_type (Ident.create "string")
 
-let id_couple = Ident.create "couple"
-
-let couple a b = logic_type ~args: [ a; b ] id_couple
-
-let rec tuple_of couple a i j =
-  let c = j - i in
-  if c <= 1 then
-    a.(i)
-  else
-    let m = i + c / 2 in
-    couple
-      (tuple_of couple a i m)
-      (tuple_of couple a m j)
-
-let make_tuple unit couple = function
-  | [] ->
-      unit
-  | l ->
-      let a = Array.of_list l in
-      let c = Array.length a in
-      tuple_of couple a 0 c
-
-let rec make_proj fst snd n i x =
-  if n <= 1 then x else
-    let m = n / 2 in
-    if i < m then
-      make_proj fst snd m i (fst x)
-    else
-      make_proj fst snd (n - m) (i - m) (snd x)
-
-let tuple = make_tuple unit couple
+let tuple l = LTTuple l
 
 type term =
   | LVoid
   | LInt of string
   | LReal of string
   | LBool of bool
+  | LTuple of term list
+  | LProj of int * int * term
   | LApp of Ident.t * term list
   | LVar of Ident.t
   | LDeref of Ident.t * Ident.t option
@@ -116,6 +89,7 @@ type term =
   | LGe of term * term
 
 let at_old = Ident.create ""
+let at_init = Ident.create "Init"
 
 let term_void = LVoid
 let term_ints x = LInt x
@@ -235,6 +209,8 @@ type expr =
   | EAbsurd
   | ELet of Ident.t * expr * expr
   | EAssign of Ident.t * expr
+  | ETuple of expr list
+  | EProj of int * int * expr
 
 let rec may_have_side_effects = function
   | EVoid | EInt _ | EReal _ | EBool _ | EVar _ | EDeref _ ->
@@ -267,6 +243,8 @@ let rec may_have_side_effects = function
       || (List.exists may_have_side_effects l)
   | ESeq l ->
       List.exists may_have_side_effects l
+  | ETuple l -> List.exists may_have_side_effects l
+  | EProj (_, _, e) -> may_have_side_effects e
   | EApp _ | ERaise _ | EFun _ | EAssert _ | ELabel _ | EBlackBox _ | EAbsurd
   | EAssign _ ->
       true
@@ -377,7 +355,7 @@ let rec seq = function
         | y ->
             ESeq [ x; y ]
       end
-      
+
 let seq2 x y = seq [ x; y ]
 let app x y = EApp (x, y)
 let app_list x l = List.fold_left app x l
@@ -393,23 +371,6 @@ let absurd = EAbsurd
 let expr_let id x y = ELet (id, x, y)
 let assign id x = EAssign (id, x)
 
-let pair, expr_pair =
-  let id = Ident.create "pair" in
-  (fun a b -> term_app id [ a; b ]),
-  (fun a b -> app (app (var id) a) b)
-
-let fst, expr_fst =
-  let id = Ident.create "fst" in
-  (fun a -> term_app id [ a ]),
-  (fun a -> app (var id) a)
-
-let snd, expr_snd =
-  let id = Ident.create "snd" in
-  (fun a -> term_app id [ a ]),
-  (fun a -> app (var id) a)
-
-let proj = make_proj fst snd
-
 let expr_ignore =
   let underscore = Ident.create "_" in
   fun a b ->
@@ -418,11 +379,11 @@ let expr_ignore =
     else
       b
 
-let term_tuple = make_tuple term_void pair
-let term_proj = make_proj fst snd
+let term_tuple l = LTuple l
+let term_proj n i t = LProj (n, i, t)
 
-let expr_tuple = make_tuple const_void expr_pair
-let expr_proj = make_proj expr_fst expr_snd
+let expr_tuple l = ETuple l
+let expr_proj n i e = EProj (n, i, e)
 
 let assume = function
   | ABool true ->
@@ -436,14 +397,14 @@ type decl =
   | DParam of Ident.t * typ
   | DLet of Ident.t * expr
   | DPredicate of Ident.t * (Ident.t * logic_type) list * assertion option
-  | DAxiom of Ident.t * assertion
+  | DAxiom of bool * Ident.t * assertion
 
 let decl_logic_type ?(args = []) name = DLogicType (args, name)
 let decl_logic name ?(args = []) t = DLogic (name, args, t)
 let decl_parameter name t = DParam (name, t)
 let decl_let name e = DLet (name, e)
 let decl_predicate name args body = DPredicate (name, args, body)
-let decl_axiom name body = DAxiom (name, body)
+let decl_axiom ~is_lemma name body = DAxiom (is_lemma, name, body)
 
 type file = decl list
 
@@ -471,6 +432,53 @@ let prime_identifier fmt id =
   fprintf fmt "'%a" identifier id
 
 module Why2 = struct
+  let couple =
+    let id_couple = Ident.create "couple" in
+    fun a b -> logic_type ~args: [ a; b ] id_couple
+
+  let rec tuple_of couple a i j =
+    let c = j - i in
+    if c <= 1 then
+      a.(i)
+    else
+      let m = i + c / 2 in
+      couple
+        (tuple_of couple a i m)
+        (tuple_of couple a m j)
+
+  let make_tuple unit couple = function
+    | [] ->
+        unit
+    | l ->
+        let a = Array.of_list l in
+        let c = Array.length a in
+        tuple_of couple a 0 c
+
+  let rec make_proj fst snd n i x =
+    if n <= 1 then x else
+      let m = n / 2 in
+      if i < m then
+        make_proj fst snd m i (fst x)
+      else
+        make_proj fst snd (n - m) (i - m) (snd x)
+
+  let pair, expr_pair =
+    let id = Ident.create "pair" in
+    (fun a b -> term_app id [ a; b ]),
+    (fun a b -> app (app (var id) a) b)
+
+  let fst, expr_fst =
+    let id = Ident.create "fst" in
+    (fun a -> term_app id [ a ]),
+    (fun a -> app (var id) a)
+
+  let snd, expr_snd =
+    let id = Ident.create "snd" in
+    (fun a -> term_app id [ a ]),
+    (fun a -> app (var id) a)
+
+  let proj = make_proj fst snd
+
   let rec pp_logic_type fmt t =
     let o x = fprintf fmt x in
     match t with
@@ -484,6 +492,8 @@ module Why2 = struct
                 o "@[<hov 2>(%a)@ %a@]" (list ~sep: comma_space pp_logic_type)
                   l identifier t.lt_name
           end
+      | LTTuple l ->
+          pp_logic_type fmt (make_tuple unit couple l)
       | LTVar t ->
           o "'%a" identifier t
 
@@ -500,6 +510,8 @@ module Why2 = struct
             o "(%s)" s
       | LBool b ->
           o "%b" b
+      | LTuple l -> pp_term fmt (make_tuple term_void pair l)
+      | LProj (n, i, t) -> pp_term fmt (make_proj fst snd n i t)
       | LApp (f, args) ->
           o "@[<hv 2>%a%a@]"
             identifier f
@@ -752,6 +764,10 @@ module Why2 = struct
           o "(@[%a >=@ %a@])" pp_expr x pp_expr y
       | EEq (x, y) ->
           o "(@[%a =@ %a@])" pp_expr x pp_expr y
+      | ETuple l -> 
+          pp_expr fmt (make_tuple const_void expr_pair l)
+      | EProj (n, i, e) ->
+          pp_expr fmt (make_proj expr_fst expr_snd n i e)
 
   let pp_decl fmt d =
     let o x = fprintf fmt x in
@@ -788,8 +804,14 @@ module Why2 = struct
                ~sep: comma_space
                (fun fmt (_, ty) ->
                   fprintf fmt "%a" pp_logic_type ty)) args
-      | DAxiom (name, body) ->
-          o "@[<hv 2>axiom %a:@ %a@]@.@." identifier name pp_assertion body
+      | DAxiom (is_lemma, name, body) ->
+	  if is_lemma then
+	    begin
+              o "@[<hv 2>goal %a_as_lemma:@ %a@]@.@."
+                identifier name pp_assertion body
+	    end;
+          o "@[<hv 2>axiom %a:@ %a@]@.@." 
+	    identifier name pp_assertion body
 
   let to_channel ch file =
     let fmt = formatter_of_out_channel ch in
@@ -802,17 +824,15 @@ module Why3 = struct
   let rec pp_logic_type fmt t =
     let o x = fprintf fmt x in
     match t with
-(*
-      | LTNormal t when t.lt_name == id_couple ->
+      | LTTuple l ->
           o "@[<hov 2>(%a)@]"
-            (list ~sep:comma_space pp_logic_type) t.lt_args          
-*)
+            (list ~sep: comma_space pp_logic_type) l
       | LTNormal t ->
           begin match t.lt_args with
             | [] ->
                 o "%a" identifier t.lt_name
             | l ->
-                o "@[<hov 2>(%a@ %a)@]"
+                o "@[<hov 2>%a@ (%a)@]"
                   identifier t.lt_name
                   (list ~sep: space pp_logic_type) l
           end
@@ -832,6 +852,14 @@ module Why3 = struct
             o "(%s)" s
       | LBool b ->
           o "%s" (if b then "True" else "False")
+      | LTuple l ->
+          o "(%a)" (list ~sep: comma_space pp_term) l
+      | LProj (n, i, t) ->
+          o "@[<hv 2>(let (";
+          for j = 1 to i do o "_," done;
+          o "x";
+          for j = i+1 to n-1 do o ",_" done;
+          o ")@ = %a in x)@]" pp_term t
       | LApp (f, args) ->
           o "@[<hv 2>(%a%a)@]"
             identifier f
@@ -847,7 +875,7 @@ module Why3 = struct
           if Ident.name at = "" then (* hack for old *)
             o "@[<hv 2>(old@ !%a)@]" identifier x
           else
-            o "@[<hv 2>(at@ !%a@ %a)@]" identifier x identifier at
+            o "@[<hv 2>(at@ !%a@ %a)@]" identifier x capitalized_identifier at
       | LIf (a, b, c) ->
           o "(@[<hv>@[<hv 2>if %a = True then@ %a@]@ @[<hv 2>else@ %a@]@])"
             pp_term a pp_term b pp_term c
@@ -941,7 +969,7 @@ module Why3 = struct
       | TBase lt ->
           o "%a" pp_logic_type lt
       | TRef t ->
-          o "@[<hov 2>ref@ %a@]" pp_type t
+          o "@[<hov 2>ref@ (%a)@]" pp_type t
       | TFun (id, pt, rt) ->
           o "@[<hov 2>%a:@ %a ->@ %a@]" identifier id pp_type pt pp_type rt
       | TAnnot (pre, reads, writes, post, exceptions, t) ->
@@ -1053,7 +1081,7 @@ module Why3 = struct
       | EAssert a ->
           o "@[<hv 2>assert@ { %a }@]" pp_assertion a
       | ELabel (x, e) ->
-          o "(@[%a:@ %a@])" identifier x pp_expr e
+          o "(@[label %a :@ %a@])" capitalized_identifier x pp_expr e
       | EBlackBox (TRef t) ->
           o "@[ref (any %a)@]" pp_type t
       | EBlackBox t ->
@@ -1076,6 +1104,14 @@ module Why3 = struct
           o "(@[%a >=@ %a@])" pp_expr x pp_expr y
       | EEq (x, y) ->
           o "(@[%a =@ %a@])" pp_expr x pp_expr y
+      | ETuple l ->
+          o "(%a)" (list ~sep: comma_space pp_expr) l
+      | EProj (n, i, e) ->
+          o "@[<hv 2>(let (";
+          for j = 1 to i do o "_," done;
+          o "x";
+          for j = i+1 to n-1 do o ",_" done;
+          o ")@ = %a in x)@]" pp_expr e
 
   let pp_decl fmt d =
     let o x = fprintf fmt x in
@@ -1092,7 +1128,8 @@ module Why3 = struct
       | DLogic (name, args, res) ->
           o "{ @[<hov 2>logic %a%a:@ %a@] }@.@."
             identifier name
-            (list ~left: space ~sep: space pp_logic_type) args
+            (list ~left: space ~sep: space
+               (fun fmt t -> fprintf fmt "(%a)" pp_logic_type t)) args
             pp_logic_type res
       | DParam (name, t) ->
           o "@[<hov 2>parameter %a:@ %a@]@.@."
@@ -1110,8 +1147,9 @@ module Why3 = struct
                   fprintf fmt "(%a: %a)" identifier id pp_logic_type ty)) args
             (option ~left: (fun fmt () -> fprintf fmt " =@ ") pp_assertion)
             body
-      | DAxiom (name, body) ->
-          o "{ @[<hv 2>axiom %a:@ %a@]@ }@]@.@."
+      | DAxiom (is_lemma, name, body) ->
+          o "{ @[<hv 2>%s %a:@ %a@]@ }@]@.@."
+	    (if is_lemma then "lemma" else "axiom")
             capitalized_identifier name pp_assertion body
 
   let to_channel ch file =
